@@ -5,7 +5,7 @@
 # Released under the MIT license. See LICENSE file for details.
 #
 
-__version__ = (0, 9, 0)
+__version__ = (0, 9, 1)
 
 import sys
 import socket
@@ -46,15 +46,15 @@ CHAINSFILE_IPV6 = os.path.join(BASEDIR, "ipv6.rules")
 
 
 rule_explanation = """\
-DIRECTION ACTION STATES PROTOCOL OPTIONS SOURCE DESTINATION LOG [LOGLEVEL] [LOGNAME]
+DIRECTION ACTION [STATES] PROTOCOL OPTIONS SOURCE DESTINATION LOG [LOGLEVEL] [LOGNAME]
 
 DIRECTION   := ["local"] "in" | "out"
 ACTION      := "permit" | "deny" | "snat" NATARGS | "dnat" NATARGS | "masq"
 STATES      := "NONE" | STATE[,STATE ...]
-STATE       := "ESTABLISHED" | "NEW" | "RELATED" | "INVALID"
+STATE       := "NEW" | "ESTABLISHED" | "RELATED" | "INVALID"
 PROTOCOL    := "ip" | "all" | "tcp" | "udp" | "icmp" | number | `/etc/protocol`
 DESTINATION := SOURCE
-ADDR        := group | hostname | ip/mask | "any"
+ADDR        := group | fqdn-hostname | ip/mask | "any"
 PORT        := number | "all"
 LOG         := log [syslog_level]
 
@@ -81,7 +81,7 @@ SOURCE      := ADDR PORT
 OPTIONS     :=
 """
 
-defaults_txt = """\
+rule_defaults_txt = """\
 Default tcp state: NEW
 """
 
@@ -344,12 +344,12 @@ class Group(list):
                 for ip in ipv4:
                     if not ip in self.ipv4:
                         if ip.network != ip.ip:
-                            raise FWIPMaskBoundaryError(ip)
+                            raise FWIPMaskBoundaryError(ip, self.lineno)
                         self.ipv4.append(ip)
                 for ip in ipv6:
                     if not ip in self.ipv6:
                         if ip.network != ip.ip:
-                            raise FWIPMaskBoundaryError(ip)
+                            raise FWIPMaskBoundaryError(ip, self.lineno)
                         self.ipv6.append(ip)
         self.resolved = True
         self.ipv4.sort()
@@ -360,6 +360,7 @@ class Group(list):
         assert self.resolved is not None
         return self.ipv4 + self.ipv6
 
+
 class Hostname(Group):
 
     def resolve(self):
@@ -544,11 +545,11 @@ class FWPreprocess(Scanner):
         (Str("udp"),        udp_action),
         (Str("icmp"),       icmp_action),
         (resword,           TEXT),
+        (number,            "number"),
         (ip4,               "ip4"),
         (ip6,               "ip6"),
         (name,              "name"),
         (hostname,          "hostname"),
-        (number,            "number"),
         (stringlit,         "string"),
         (punctuation,       TEXT),
         (lineterm,          newline_action),
@@ -916,9 +917,13 @@ class FWPreprocess(Scanner):
         else:
             invert = ""
         if not token in ["ip", "all", "icmp", "tcp", "udp", "number"]:
-            self.log_error("Invalid protocol '%s'" % text)
-            token, text = self.skip_this_line()
-            return None
+            if text in self.protocols:
+                token = "number"
+                text = self.protocols[text]
+            else:
+                self.log_error("Invalid protocol '%s'" % text)
+                token, text = self.skip_this_line()
+                return None
         if rule.action in ["dnat"]:
             if token in ["tcp", "udp"]:
                 if not rule.natports:
@@ -1262,7 +1267,7 @@ class FWPreprocess(Scanner):
             try:
                 ip = netaddr.IPNetwork(name)
                 if ip.network != ip.ip:
-                    self.log_error(FWIPMaskBoundaryError(ip).log_message())
+                    self.log_error(FWIPMaskBoundaryError(ip, rule.lineno).log_message())
                 ips = [ip]
             except netaddr.core.AddrFormatError, e:
                 if self.groups.has_key(name) and \
@@ -1394,7 +1399,8 @@ class FWPreprocess(Scanner):
             except FWResolveError, e:
                 self.log_error(e.log_message(), e.lineno)
         for group in self.groups.itervalues():
-            if group.lineno is None:
+            if group.lineno is None and not \
+               isinstance(group, Hostname):
                 raise FWUndefinedGroup(group.name, group.referred_lines[0])
 
     def make_rule(self, chainnr, chainname, iface, rule):
@@ -1407,10 +1413,11 @@ class FWPreprocess(Scanner):
         dsts_ip4, dsts_ip6 = self.resolve_ip(rule.destinations, rule)
         lines_ip4 = []
         lines_ip6 = []
-        line = []
+        line_ipv4 = []
+        line_ipv6 = []
         targets = []
         if rule.nat or rule.action == "masq":
-            line += ["-t nat"]
+            line_ipv4 += ["-t nat"]
             if not srcs_ip4 or not dsts_ip4:
                 self.log_error("NAT rule only valid for IPv4", rule.lineno)
             else:
@@ -1422,7 +1429,8 @@ class FWPreprocess(Scanner):
                     if dst[1] != all:
                         self.log_warning("Ignoring %s rule for IPv6 destination address %s" % (rule.action, dst), rule.lineno)
         else:
-            line += ["-t filter"]
+            line_ipv4 += ["-t filter"]
+            line_ipv6 += ["-t filter"]
         if rule.logging:
             lineno = rule.lineno
             action = rule.action
@@ -1435,16 +1443,24 @@ class FWPreprocess(Scanner):
             # iptables-restore needs strings in "" and not ''
             targets.append('LOG --log-prefix "%s " --log-level %s -m limit --limit 60/minute --limit-burst 10' % (s, rule.logging))
         if rule.direction == "in":
-            line += ["-i", iface]
+            line_ipv4 += ["-i", iface]
+            line_ipv6 += ["-i", iface]
         elif rule.direction == "out":
-            line += ["-o", iface]
+            line_ipv4 += ["-o", iface]
+            line_ipv6 += ["-o", iface]
         else:
             self.log_error("Invalid direction '%s'" % rule.direction, rule.lineno)
         chainname = rule.chainname(chainnr, chainname, iface)
-        line += ["-p", rule.protocol]
+        line_ipv4 += ["-p", rule.protocol]
+        if rule.protocol == 'icmp':
+            line_ipv6 += ["-p", 'icmpv6']
+        else:
+            line_ipv6 += ["-p", rule.protocol]
         if rule.state:
-            line += ["-m state --state", rule.state]
-        line += ["-A %d%s" % (100 + chainnr, chainname)]
+            line_ipv4 += ["-m state --state", rule.state]
+            line_ipv6 += ["-m state --state", rule.state]
+        line_ipv4 += ["-A %d%s" % (100 + chainnr, chainname)]
+        line_ipv6 += ["-A %d%s" % (100 + chainnr, chainname)]
         if rule.nat:
             if rule.natports:
                 nat = "%s:%s" % (rule.nat, rule.natports)
@@ -1452,10 +1468,10 @@ class FWPreprocess(Scanner):
                 nat = rule.nat
             if rule.action == "snat":
                 targets.append("SNAT")
-                line += ["-j %(target)s --to-source", nat]
+                line_ipv4 += ["-j %(target)s --to-source", nat]
             else:
                 targets.append("DNAT")
-                line += ["-j %(target)s --to-destination", nat]
+                line_ipv4 += ["-j %(target)s --to-destination", nat]
         else:
             if rule.action == "permit":
                 if rule.direction == "in" and \
@@ -1463,14 +1479,18 @@ class FWPreprocess(Scanner):
                     targets.append("RETURN")
                 else:
                     targets.append("ACCEPT")
-                line += ["-j %(target)s"]
+                line_ipv4 += ["-j %(target)s"]
+                line_ipv6 += ["-j %(target)s"]
             elif rule.action == "deny":
                 targets.append("DROP")
-                line += ["-j %(target)s"]
+                line_ipv4 += ["-j %(target)s"]
+                line_ipv6 += ["-j %(target)s"]
             elif rule.action == "masq":
                 targets.append("MASQUERADE")
-                line += ["-j %(target)s"]
-        line_start = " ".join(line)
+                line_ipv4 += ["-j %(target)s"]
+                line_ipv6 += ["-j %(target)s"]
+        line_ipv4_start = " ".join(line_ipv4)
+        line_ipv6_start = " ".join(line_ipv6)
         # Get all src ports (two lists: ranges and comma sep)
         src_comma_ports, src_range_ports = self.resolve_ports(rule.srcports, rule)
         # Get all destination ips
@@ -1485,8 +1505,10 @@ class FWPreprocess(Scanner):
             destinations = dsts_ip4 + dsts_ip6
         for src_invert, src_ip in sources:
             if src_ip.version == 4:
+                line_start = line_ipv4_start
                 lines = lines_ip4
             else:
+                line_start = line_ipv6_start
                 lines = lines_ip6
             if src_ip.prefixlen == 0:
                 src = ""
@@ -1862,7 +1884,7 @@ usage: %%prog [options] FILE
 Rule format:
 %(rule_explanation)s
 Defaults:
-%(defaults_txt)s
+%(rule_defaults_txt)s
 ICMP options:
 %(icmp_options_txt)s
 """ % globals(),
@@ -1915,9 +1937,13 @@ ICMP options:
     fwprepocess.chainsdir_ip6 = opts.chainsdir_ip6
     fwprepocess.logtag = opts.logtag
 
-    fwprepocess.read_fwrules()
-    fwprepocess.resolve()
-    chains4, chains6 = fwprepocess.make_rules()
+    try:
+        fwprepocess.read_fwrules()
+        fwprepocess.resolve()
+        chains4, chains6 = fwprepocess.make_rules()
+    except FWMacroException, e:
+        sys.stderr.write(e.log_message())
+        sys.exit(1)
     if fwprepocess.nerrors == 0:
         fwprepocess.write_rules(chains4, chains6)
     else:
